################################################################################
# some backend tests

use Test::More "no_plan";
BEGIN { use_ok('Weigel::Search') }
################################################################################


use strict;
use warnings;

use Bio::Root::Exception;
use Bio::Perl;
use Error qw(:try);

# the number of files with assume in the data directory after
# database generation. 
my %backend_filecnt = ( Agrep => 5, Vmatch => 16, Hypa => 31 );

# the hits in Test.fasta we assume with n mismatches
my %hits = (
   At1g53160 => 2,
   At2g42200 => 1,
   At1g27360 => 1,
   At3g47170 => 3,
   At1g22000 => 4,
   At1g09950 => 5
);

foreach my $backendname ( keys %backend_filecnt ) {
   my $search_obj = Weigel::Search->new($backendname);
   my $sbe        = $search_obj->backend;

   # define own tmppath, so that we can check if all temporary files are deleted
   $sbe->settings->tmppath('t/tmp');
   $sbe->settings->datapath('t/data');
   mkdir("t/tmp");

   # delete everything in this directory, otherwise we fail the test
   unlink <t/tmp/*>;
   mkdir("t/data");

   $sbe->generate_database_out_of_fastafile( 't/Test.fasta',
      'Description for Test.fasta' );

# test if all files are there after database generation
################################################################################
   opendir( DIR, "t/data" ) || die "can't opendir t/data: $!";
   my @files = grep { /Test/ && -f "t/data/$_" } readdir(DIR);
   closedir DIR;

   foreach my $file (@files) {
      if ( $file =~ /nfo$/ ) {
         open( FILE, "t/data/$file" );
         my $filecontent = '';
         while (<FILE>) {
            $filecontent .= $_;
         }
         close(FILE);
         ok( $filecontent eq 'Description for Test.fasta', 'Description ok' );
      }
   }
   ok( scalar(@files) == $backend_filecnt{$backendname},
      "$backend_filecnt{$backendname} $backendname index files created" );

# now search for 1 to 5 mismatches, test if reverse complement works
################################################################################
   $sbe->settings->database('Test.fasta');
   for my $j ( 0 .. 1) {
      my $sequence = 'tgacagaagagagtgagcac'; 
      if ($j == 0 ) {
         $sbe->settings->query($sequence);
         $sbe->settings->reverse_complement(1);
      } else {
         $sbe->settings->query(revcom_as_string($sequence));
         $sbe->settings->reverse_complement(0);
      }
      for my $i ( 1 .. 5 ) {
         $sbe->settings->mismatches($i);

         try {
            $sbe->search();
           } catch Bio::Root::SystemException with {
            my $E = shift;
              print STDERR 'Backend call failed: '
              . $E->{'-text'} . ' ('
              . $E->{'-line'} . ")\n";
           } otherwise {
            my $E = shift;
              print STDERR "An unexpected exception occurred: \n$E";
           };
         my @ids = ();
         foreach my $res ( @{ $sbe->results } ) {
          #   print STDERR "IS: " . $res->sequence->id . "\n";
            push ( @ids, $res->sequence->id );
         }
         my @shouldbe = _get_ids_where_mm_smaller($i);
         foreach (@shouldbe) {
         #    print "SHOULDBE: $_\n";
         }
         @ids = sort @ids;
         ok( compare_arrays( \@shouldbe, \@ids ), "Results ok" );
      }
   }

# testing, if backends find matches at the end of long sequences
################################################################################
   $sbe->settings->query('CGAGCTGATGCAAAGCTCGCGGGACTGA');
   $sbe->settings->reverse_complement(0);
   $sbe->settings->mismatches(0);
   try {
      $sbe->search();
     } catch Bio::Root::SystemException with {
      my $E = shift;
        print STDERR 'Backend call failed: '
        . $E->{'-text'} . ' ('
        . $E->{'-line'} . ")\n";
     } otherwise {
      my $E = shift;
        print STDERR "An unexpected exception occurred: \n$E";
     };
   my @ids = ();
   foreach my $res ( @{ $sbe->results } ) {
      push ( @ids, $res->sequence->id );
   }
   my @shouldbe = qw( At1g67120 );
   ok( compare_arrays( \@shouldbe, \@ids ), "Results ok" );
################################################################################

   # clean up
   foreach my $file (@files) {
      unlink("t/data/$file");
   }

   ok( rmdir("t/tmp"), "Can remove tmp directory (all temp files deleted)" );
   ok( rmdir("t/data"),
"Can remove data directory (all data files deleted-just a test for the test)"
   );
}



# some helper functions
################################################################################
################################################################################
sub _get_ids_where_mm_smaller {
   my $mm = shift;
   my @results;
   foreach my $key ( keys %hits ) {
      push ( @results, $key ) if $hits{$key} <= $mm;
   }
   return sort @results;
}

sub compare_arrays {
   my ( $first, $second ) = @_;
   no warnings;    # silence spurious -w undef complaints
   return 0 unless @$first == @$second;
   for ( my $i = 0 ; $i < @$first ; $i++ ) {
      return 0 if $first->[$i] ne $second->[$i];
   }
   return 1;
}

# vim: ft=perl sw=3 ts=3 expandtab
